import rclpy
from rclpy.node import Node, SetParametersResult
from typing import Union, List, Tuple, Any
from rcl_interfaces.msg import Parameter, ParameterValue
from rcl_interfaces.srv import SetParametersAtomically, ListParameters, GetParameters
from dataclasses import dataclass

from lifecycle_msgs.msg import State
from lifecycle_msgs.msg import Transition
from ble_controller.lifecycle import LifecycleNode

from lifecycle_msgs.msg import State
from lifecycle_msgs.msg import Transition

from lifecycle_msgs.srv import ChangeState
from lifecycle_msgs.srv import GetAvailableTransitions
from lifecycle_msgs.srv import GetState


import os
from os.path import join, dirname, realpath
from ament_index_python.packages import *
from launch import LaunchDescription, LaunchService
from launch.actions import IncludeLaunchDescription
from launch.launch_description_sources import PythonLaunchDescriptionSource

import multiprocessing

from ble_controller.utils import get_label, transition_cls, state_cls, cli_call


@dataclass
class Config:
    enable: bool


class BaseController(LifecycleNode):
    def __init__(self, name):
        super().__init__(name)
        self.declare_parameters(
            namespace="",
            parameters=[
                ("enable", None),
            ],
        )
        self.config = None
        self.logger = self.get_logger()
        self.add_on_set_parameters_callback(self.config_update)

    def on_configure(self):
        self.logger.info("on_configure() is called")
        self.config = (
            Config(
                enable=self.get_parameter("enable").value,
            )
            if self.config is None
            else self.config
        )
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def on_cleanup(self):
        self.logger.info("on_cleanup() is called")
        self.stop()
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def on_activate(self):
        self.logger.info("on_activate() is called")
        # RTSP
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def on_deactivate(self):
        self.logger.info("on_deactivate() is called")
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def destroy_node(self) -> bool:
        self.stop()
        return super().destroy_node()

    def stop(self):
        self.logger.info("Stopped")

    def config_update(self, params: List[Parameter]):
        self.logger.info(f"Update: {[str(p.name) for p in params]}")
        for param in params:
            if param.name == "enable":
                self.config.enable = param.value
            else:
                self.logger.info(f"Unknown parameter: {param.name}")
        return SetParametersResult(successful=True)

    def start(self):
        self.logger.info("Started")

    def find_nodes(self):
        self.logger.info("Finding nodes...")

        node_names = self.get_node_names()
        self.logger.info(f"Found nodes: {node_names}")
        # get unique node list
        node_list = list(set(node_names))
        return node_list

    def node_get_parameters(self, node_name) -> Union[List[Parameter], None]:
        """
        node_name: string
        """
        self.logger.info("Getting parameters...")
        cli = self.create_client(ListParameters, f"{node_name}/list_parameters")
        try:
            list_res: ListParameters.Response = cli_call(
                self, cli, ListParameters.Request(), 1.0
            )
            params_names = list_res.result.names
        except Exception as e:
            self.logger.info(f"Failed to retrieve parameter: {e}")
            return []

        cli = self.create_client(GetParameters, f"{node_name}/get_parameters")
        try:
            res: GetParameters.Response = cli_call(
                self, cli, GetParameters.Request(names=params_names), 1.0
            )
            params_values: list[ParameterValue] = res.values
        except Exception as e:
            self.logger.info(f"Failed to retrieve parameter: {e}")
            return []

        params_names_values = [
            Parameter(name=params_names[i], value=params_values[i])
            for i in range(len(params_names))
        ]
        return params_names_values

    def node_change_parameter(self, node_name, list_of_parameters: List[Parameter]):
        """
        node_name: string
        list_of_parameters: list of parameters: [Parameter, Parameter, ...)]
        """
        self.logger.info("Changing parameters...")
        cli = self.create_client(
            SetParametersAtomically, f"{node_name}/set_parameters_atomically"
        )
        try:
            res: SetParametersAtomically.Response = cli_call(
                self,
                cli,
                SetParametersAtomically.Request(parameters=list_of_parameters),
                1.0,
            )
            return True if res is not None else False
        except Exception as e:
            self.logger.info(f"Failed to change parameter: {e}")
            return False

    def node_get_state(self, node_name) -> Union[State, None]:
        self.logger.info(f"Getting state for {node_name}...")
        cli = self.create_client(GetState, f"{node_name}/get_state")
        try:
            res: GetState.Response = cli_call(self, cli, GetState.Request(), 3.0)
            return res.current_state
        except Exception as e:
            self.logger.info(f"Failed to get state for {node_name}: {e}")
            return None

    def node_get_available_transitions(
        self, node_name
    ) -> Union[List[Transition], None]:
        self.logger.info(f"Getting available transitions for {node_name}...")
        cli = self.create_client(
            GetAvailableTransitions, f"{node_name}/get_available_transitions"
        )
        try:
            res: GetAvailableTransitions.Response = cli_call(
                self, cli, GetAvailableTransitions.Request(), 3.0
            )
            return res.available_transitions
        except Exception as e:
            self.logger.info(f"Failed to get available transitions for {node_name}")
            return None

    def node_change_state(self, node_name, transition: Transition) -> bool:
        self.logger.info(f"Changing state for {node_name}...")
        cli = self.create_client(ChangeState, f"{node_name}/change_state")
        try:
            res: ChangeState.Response = cli_call(
                self, cli, ChangeState.Request(transition=transition), 5.0
            )
            return True if res is not None else False
        except Exception as e:
            self.logger.info(f"Failed to change state for {node_name}: {e}")
            return False

    def node_restart(self, node_name) -> bool:
        self.logger.info(f"Restarting {node_name}...")
        node_state = self.node_get_state(node_name)
        if node_state is None:
            return False
        if node_state.id == State.PRIMARY_STATE_ACTIVE:
            chain = [
                Transition.TRANSITION_DEACTIVATE,
                Transition.TRANSITION_CLEANUP,
                Transition.TRANSITION_CONFIGURE,
                Transition.TRANSITION_ACTIVATE,
            ]
        elif node_state.id == State.PRIMARY_STATE_INACTIVE:
            chain = [
                Transition.TRANSITION_CLEANUP,
                Transition.TRANSITION_CONFIGURE,
                Transition.TRANSITION_ACTIVATE,
            ]
        elif node_state.id == State.PRIMARY_STATE_UNCONFIGURED:
            chain = [Transition.TRANSITION_CONFIGURE, Transition.TRANSITION_ACTIVATE]
        else:
            self.logger.info(f"Unknown state for restarting {node_name}")
            return False

        for transition in chain:
            success = self.node_change_state(node_name, transition_cls(transition))
            if not success:
                return False
        return True

    def node_reconfigure(self, node_name) -> bool:
        self.logger.info(f"Reconfiguring {node_name}...")
        node_state = self.node_get_state(node_name)
        if node_state is None:
            return False
        if node_state.id == State.PRIMARY_STATE_ACTIVE:
            chain = [Transition.TRANSITION_DEACTIVATE, Transition.TRANSITION_ACTIVATE]
        elif node_state.id == State.PRIMARY_STATE_INACTIVE:
            chain = [Transition.TRANSITION_ACTIVATE]
        elif node_state.id == State.PRIMARY_STATE_UNCONFIGURED:
            chain = [Transition.TRANSITION_CONFIGURE, Transition.TRANSITION_ACTIVATE]
        else:
            self.logger.info(f"Unknown state for Reconfiguring {node_name}")
            return False

        for transition in chain:
            success = self.node_change_state(node_name, transition_cls(transition))
            if not success:
                return False
        return True

    def node_destroy(self, node_name) -> bool:
        self.logger.info(f"Destroying {node_name}...")
        node_state = self.node_get_state(node_name)
        if node_state is None:
            return False
        if node_state.id == State.PRIMARY_STATE_ACTIVE:
            chain = [
                Transition.TRANSITION_DEACTIVATE,
                Transition.TRANSITION_CLEANUP,
                Transition.TRANSITION_UNCONFIGURED_SHUTDOWN,
                Transition.TRANSITION_DESTROY,
            ]
        elif node_state.id == State.PRIMARY_STATE_INACTIVE:
            chain = [
                Transition.TRANSITION_CLEANUP,
                Transition.TRANSITION_UNCONFIGURED_SHUTDOWN,
                Transition.TRANSITION_DESTROY,
            ]
        elif node_state.id == State.PRIMARY_STATE_UNCONFIGURED:
            chain = [
                Transition.TRANSITION_UNCONFIGURED_SHUTDOWN,
                Transition.TRANSITION_DESTROY,
            ]
        else:
            self.logger.info(f"Unknown state for destroying {node_name}")
            return False

        for transition in chain:
            success = self.node_change_state(node_name, transition_cls(transition))
            if not success:
                return False
        return True

    def packages_get(self) -> List[str]:
        self.logger.info("Getting ros2 packages...")
        packages = get_packages_with_prefixes()
        return sorted(list(packages.keys()))

    def packages_get_launch_files(self, package_name) -> List[str]:
        self.logger.info(f"Getting launch files for {package_name}...")
        package_path = get_package_share_directory(package_name)
        launch_files = []
        if not os.path.exists(join(package_path, "launch")):
            return launch_files
        for file in os.listdir(join(package_path, "launch")):
            if file.endswith(".launch.py"):
                launch_files.append(file)
        return launch_files

    def packages_launch(self, package_name, launch_file_name) -> bool:
        self.logger.info(f"Launching {launch_file_name} from {package_name}...")
        try:
            os.system(
                f"nohup ros2 launch {package_name} {launch_file_name} > /dev/null 2>&1 &"
            )
        except Exception as e:
            self.logger.info(f"Failed to launch {launch_file_name} from {package_name}")
            return False
        return True

    def restart_all(self):
        self.logger.info("Restarting all nodes...")
        try:
            os.system("systemctl restart ros2_nodes.service")
        except Exception as e:
            self.logger.info(f"Failed to restart all nodes: {e}")
            return False
        return True
    

    def restart_controller(self):
        self.logger.info("Restarting all nodes...")
        try:
            os.system("systemctl restart ros2_nodes_controller.service")
        except Exception as e:
            self.logger.info(f"Failed to restart all nodes: {e}")
            return False
        return True
    ## Abstact input UI:
    # When user press scan button, call find_nodes()
    # When user press the node name, call get_node_parameters(node_name)
    # When user press change button, after the user change the value, call change_parameter(node_name, list_of_parameters), then call get_node_parameters(node_name) to check if the change is successful
