import rclpy
from rcl_interfaces.msg import Parameter
from ble_controller.utils import str_to_value2, extract_value
from ble_controller.base_controller import BaseController
import flask
from flask import Response
import json
from lifecycle_msgs.msg import State
from lifecycle_msgs.msg import TransitionDescription
from typing import List

from ble_controller.utils import (
    get_label,
    transition_cls,
    state_cls,
)


class WebController(BaseController):
    def __init__(self):
        super().__init__("web_controller")
        self.app = flask.Flask(__name__)
        self.app.config["DEBUG"] = True
        self.app.add_url_rule("/", "index", self.index)
        self.app.add_url_rule("/scan", "scan", self.scan)
        
        self.app.add_url_rule("/parameters", "parameters", self.parameters)
        self.app.add_url_rule("/parameters", "update_parameters", self.update, methods=["POST"])
        
        self.app.add_url_rule("/state", "states", self.web_get_state)
        self.app.add_url_rule("/transitions", "transitions", self.web_get_transitions)
        self.app.add_url_rule("/transitions", "transit", self.web_change_state, methods=["POST"])
        self.app.add_url_rule("/restart", "restart", self.web_restart)
        self.app.add_url_rule("/reconfig", "reconfig", self.web_reconfig)
        self.app.add_url_rule("/shutdown", "shutdown", self.web_destroy)
        
        self.app.add_url_rule("/packages", "packages", self.web_get_packages)
        self.app.add_url_rule("/launch_files", "launch_files", self.web_get_launch_files)
        self.app.add_url_rule("/launch", "launch", self.web_launch, methods=["POST"])
        
        self.app.add_url_rule("/restart_all", "restart_all", self.web_restart_all)
        self.app.add_url_rule("/restart_controller", "restart_controller", self.web_restart_controller)
        
    def start(self):
        super().start()
        self.app.run(host="horizon", port=7000)

    def destroy_node(self) -> bool:
        self.app = None
        return super().destroy_node()
    
    def index(self):
        return Response("WebController", status=200)

    def scan(self):
        node_list = self.find_nodes()
        node_list.sort()
        return Response(json.dumps(node_list), status=200)

    def parameters(self):
        node_name = flask.request.args.get("node_name")
        if node_name is None:
            return Response("node_name is not provided", status=400)
        params = self.node_get_parameters(node_name)
        return_params = [dict(name=param.name, value=extract_value(param.value), type=param.value.type) for param in params]
        return Response(json.dumps(return_params), status=200)

    # receive form data
    def update(self):
        # params is json in the form of [{"name": "param_name", "value": "param_value", "type": "param_type"}]
        node_name = flask.request.args.get("node_name")
        params = flask.request.get_data()
        params = json.loads(params)
        if node_name is None:
            return Response("node_name is not provided", status=400)
        params = [Parameter(name=param["name"], value=str_to_value2(param["type"], param["value"])) for param in params]
        sucess = self.node_change_parameter(node_name, params)
        if not sucess:
            return Response("Unable to update", status=400)
        return Response("Updated", status=200)

    def web_get_state(self):
        node_name = flask.request.args.get("node_name")
        if node_name is None:
            return Response("node_name is not provided", status=400)
        state: State = self.node_get_state(node_name)
        if state is None:
            return Response("Unable to get state", status=400)
        return Response(json.dumps({"id": state.id, "label": state.label}), status=200)
    
    def web_get_transitions(self):
        node_name = flask.request.args.get("node_name")
        if node_name is None:
            return Response("node_name is not provided", status=400)
        transitions: List[TransitionDescription] = self.node_get_available_transitions(node_name)
        if transitions is None:
            return Response("Unable to get transitions", status=400)
        transitions = [{"id": t.transition.id, "label": t.transition.label} for t in transitions]
        return Response(json.dumps(transitions), status=200)
    
    def web_change_state(self):
        node_name = flask.request.args.get("node_name")
        if node_name is None:
            return Response("node_name is not provided", status=400)
        transition_id = flask.request.args.get("transition_id")
        if transition_id is None:
            return Response("transition_id is not provided", status=400)
        transition = transition_cls(int(transition_id))
        print(transition)
        success = self.node_change_state(node_name, transition)
        if not success:
            return Response("Unable to change state", status=400)
        return Response("Changed state", status=200)

    def web_restart(self):
        node_name = flask.request.args.get("node_name")
        if node_name is None:
            return Response("node_name is not provided", status=400)
        success = self.node_restart(node_name)
        if not success:
            return Response("Unable to restart", status=400)
        return Response("Restarted", status=200)
    
    def web_reconfig(self):
        node_name = flask.request.args.get("node_name")
        if node_name is None:
            return Response("node_name is not provided", status=400)
        success = self.node_reconfigure(node_name)
        if not success:
            return Response("Unable to reconfig", status=400)
        return Response("reconfiged", status=200)
    
    def web_destroy(self):
        node_name = flask.request.args.get("node_name")
        if node_name is None:
            return Response("node_name is not provided", status=400)
        success = self.node_destroy(node_name)
        if not success:
            return Response("Unable to shutdown", status=400)
        return Response("shutdown", status=200)
    
    def web_get_packages(self):
        packages = self.packages_get()
        return Response(json.dumps(packages), status=200)
    
    def web_get_launch_files(self):
        package_name = flask.request.args.get("package_name")
        if package_name is None:
            return Response("package_name is not provided", status=400)
        launch_files = self.packages_get_launch_files(package_name)
        return Response(json.dumps(launch_files), status=200)
    
    def web_launch(self):
        package_name = flask.request.args.get("package_name")
        launch_file = flask.request.args.get("launch_file")
        if package_name is None or launch_file is None:
            return Response("package_name or launch_file is not provided", status=400)
        success = self.packages_launch(package_name, launch_file)
        if not success:
            return Response("Unable to launch", status=400)
        return Response("Launched", status=200)
    
    def web_restart_all(self):
        success = self.restart_all()
        if not success:
            return Response("Unable to restart all", status=400)
        return Response("Restarted all", status=200)
    
    def web_restart_controller(self):
        success = self.restart_controller()
        if not success:
            return Response("Unable to restart controller", status=400)
        return Response("Restarted controller", status=200)
    
        
def main(args=None):
    rclpy.init(args=args)
    controller = WebController()
    try:
        controller.start()
    except KeyboardInterrupt:
        print("KeyboardInterrupt")
    finally:
        controller.destroy_node()
        rclpy.shutdown()
